/*
 * Copyright (c) 2007 Intel Corporation. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sub license, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 * IN NO EVENT SHALL PRECISION INSIGHT AND/OR ITS SUPPLIERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#define _GNU_SOURCE 1
#include "sysdeps.h"
#include "va.h"
#include "va_backend.h"
#include "va_backend_x11.h"
#include "va_trace.h"
#include "va_fool.h"
#include "va_x11.h"
#include "va_dri.h"
#include "va_dri2.h"
#include "va_dricommon.h"
#include "va_nvctrl.h"
#include "va_fglrx.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

typedef struct {
    pthread_key_t tls_key;
} ThreadLocalState;

typedef struct {
    Display *dpy;
} ThreadLocalStateX11;

#define USE_TLS_DATA_AS_KEY \
    (sizeof(pthread_key_t) <= sizeof(((VADriverContextP)0)->tls_state))

static void
destroy_tls_state_x11(void *arg)
{
    ThreadLocalStateX11 * const state = arg;

    if (!state)
        return;

    if (state->dpy) {
        XCloseDisplay(state->dpy);
        state->dpy = NULL;
    }
    free(state);
}

ThreadLocalStateX11 *
create_tls_state_x11(Display *dpy)
{
    ThreadLocalStateX11 *tls_state;

    tls_state = calloc(1, sizeof(*tls_state));
    if (!tls_state)
        return NULL;

    tls_state->dpy = XOpenDisplay(DisplayString(dpy));
    if (!tls_state->dpy)
        goto error;
    return tls_state;

error:
    destroy_tls_state_x11(tls_state);
    return NULL;
}

static void
destroy_tls_state(void **tls_state_ptr)
{
    ThreadLocalState *tls_state = *tls_state_ptr;
    ThreadLocalStateX11 *tls_state_x11;
    pthread_key_t tls_key;

    tls_key = USE_TLS_DATA_AS_KEY ? (uintptr_t)tls_state : tls_state->tls_key;

    /* XXX: it seems the destructor is not implicitly called */
    tls_state_x11 = pthread_getspecific(tls_key);
    if (tls_state_x11) {
        destroy_tls_state_x11(tls_state_x11);
        pthread_setspecific(tls_key, NULL);
    }
    pthread_key_delete(tls_key);

    if (!USE_TLS_DATA_AS_KEY)
        free(tls_state);

    *tls_state_ptr = NULL;
}

static Bool
create_tls_state(void **tls_state_ptr, Display *dpy)
{
    ThreadLocalState *tls_state = NULL;
    pthread_key_t tls_key;

    if (pthread_key_create(&tls_key, destroy_tls_state_x11) != 0)
        return False;

    if (USE_TLS_DATA_AS_KEY)
        tls_state = (void *)(uintptr_t)tls_key;
    else {
        tls_state = malloc(sizeof(*tls_state));
        if (!tls_state)
            goto error;
        tls_state->tls_key = tls_key;
    }
    *tls_state_ptr = tls_state;
    return True;

error:
    destroy_tls_state((void **)&tls_state);
    *tls_state_ptr = NULL;
    return False;
}

Display *
va_x11_get_display(VADriverContextP ctx)
{
    ThreadLocalState * const tls_state = ctx->tls_state;
    ThreadLocalStateX11 *tls_state_x11;
    pthread_key_t tls_key;

    tls_key = USE_TLS_DATA_AS_KEY ? (uintptr_t)tls_state : tls_state->tls_key;
    tls_state_x11 = pthread_getspecific(tls_key);
    if (!tls_state_x11) {
        tls_state_x11 = create_tls_state_x11(ctx->native_dpy);
        if (!tls_state_x11)
            return NULL;
        pthread_setspecific(tls_key, tls_state_x11);
    }
    return tls_state_x11->dpy;
}

static int va_DisplayContextIsValid (
    VADisplayContextP pDisplayContext
)
{
    return (pDisplayContext != NULL && 
            pDisplayContext->pDriverContext != NULL);
}

static void va_DisplayContextDestroy (
    VADisplayContextP pDisplayContext
)
{
    VADriverContextP ctx;
    struct dri_state *dri_state;

    if (pDisplayContext == NULL)
        return;

    ctx = pDisplayContext->pDriverContext;
    dri_state = ctx->dri_state;

    if (dri_state && dri_state->close)
        dri_state->close(ctx);

    destroy_tls_state(&pDisplayContext->pDriverContext->tls_state);
    free(pDisplayContext->pDriverContext->dri_state);
    free(pDisplayContext->pDriverContext);
    free(pDisplayContext);
}


static VAStatus va_DRI2GetDriverName (
    VADisplayContextP pDisplayContext,
    char **driver_name
)
{
    VADriverContextP ctx = pDisplayContext->pDriverContext;

    if (!isDRI2Connected(ctx, driver_name))
        return VA_STATUS_ERROR_UNKNOWN;

    return VA_STATUS_SUCCESS;
}

static VAStatus va_DRIGetDriverName (
    VADisplayContextP pDisplayContext,
    char **driver_name
)
{
    VADriverContextP ctx = pDisplayContext->pDriverContext;

    if (!isDRI1Connected(ctx, driver_name))
        return VA_STATUS_ERROR_UNKNOWN;

    return VA_STATUS_SUCCESS;
}

static VAStatus va_NVCTRL_GetDriverName (
    VADisplayContextP pDisplayContext,
    char **driver_name
)
{
    VADriverContextP ctx = pDisplayContext->pDriverContext;
    Display * const x11_dpy = va_x11_get_display(ctx);
    int direct_capable, driver_major, driver_minor, driver_patch;
    Bool result;

    result = VA_NVCTRLQueryDirectRenderingCapable(x11_dpy, ctx->x11_screen,
                                                  &direct_capable);
    if (!result || !direct_capable)
        return VA_STATUS_ERROR_UNKNOWN;

    result = VA_NVCTRLGetClientDriverName(x11_dpy, ctx->x11_screen,
                                          &driver_major, &driver_minor,
                                          &driver_patch, driver_name);
    if (!result)
        return VA_STATUS_ERROR_UNKNOWN;

    return VA_STATUS_SUCCESS;
}

static VAStatus va_FGLRX_GetDriverName (
    VADisplayContextP pDisplayContext,
    char **driver_name
)
{
    VADriverContextP ctx = pDisplayContext->pDriverContext;
    Display * const x11_dpy = va_x11_get_display(ctx);
    int driver_major, driver_minor, driver_patch;
    Bool result;

    result = VA_FGLRXGetClientDriverName(x11_dpy, ctx->x11_screen,
                                         &driver_major, &driver_minor,
                                         &driver_patch, driver_name);
    if (!result)
        return VA_STATUS_ERROR_UNKNOWN;

    return VA_STATUS_SUCCESS;
}

static VAStatus va_DisplayContextGetDriverName (
    VADisplayContextP pDisplayContext,
    char **driver_name
)
{
    VAStatus vaStatus;

    if (driver_name)
	*driver_name = NULL;

    vaStatus = va_DRI2GetDriverName(pDisplayContext, driver_name);
    if (vaStatus != VA_STATUS_SUCCESS)
        vaStatus = va_DRIGetDriverName(pDisplayContext, driver_name);
    if (vaStatus != VA_STATUS_SUCCESS)
        vaStatus = va_NVCTRL_GetDriverName(pDisplayContext, driver_name);
    if (vaStatus != VA_STATUS_SUCCESS)
        vaStatus = va_FGLRX_GetDriverName(pDisplayContext, driver_name);
    return vaStatus;
}


VADisplay vaGetDisplay (
    Display *native_dpy /* implementation specific */
)
{
  VADisplay dpy = NULL;
  VADisplayContextP pDisplayContext;

  if (!native_dpy)
      return NULL;

  if (!dpy)
  {
      /* create new entry */
      VADriverContextP pDriverContext;
      struct dri_state *dri_state;
      void *tls_state;
      Bool tls_state_ok;
      pDisplayContext = calloc(1, sizeof(*pDisplayContext));
      pDriverContext  = calloc(1, sizeof(*pDriverContext));
      dri_state       = calloc(1, sizeof(*dri_state));
      tls_state_ok    = create_tls_state(&tls_state, native_dpy);
      if (pDisplayContext && pDriverContext && dri_state && tls_state_ok)
      {
	  pDisplayContext->vadpy_magic = VA_DISPLAY_MAGIC;          

	  pDriverContext->native_dpy       = native_dpy;
	  pDisplayContext->pDriverContext  = pDriverContext;
	  pDisplayContext->vaIsValid       = va_DisplayContextIsValid;
	  pDisplayContext->vaDestroy       = va_DisplayContextDestroy;
	  pDisplayContext->vaGetDriverName = va_DisplayContextGetDriverName;
          pDisplayContext->opaque          = NULL;
	  pDriverContext->dri_state 	   = dri_state;
	  pDriverContext->tls_state 	   = tls_state;
	  dpy                              = (VADisplay)pDisplayContext;
      }
      else
      {
	  if (pDisplayContext)
	      free(pDisplayContext);
	  if (pDriverContext)
	      free(pDriverContext);
          if (dri_state)
              free(dri_state);
          if (tls_state_ok)
              destroy_tls_state(&tls_state);
      }
  }
  
  return dpy;
}

#define CTX(dpy) (((VADisplayContextP)dpy)->pDriverContext)
#define CHECK_DISPLAY(dpy) if( !vaDisplayIsValid(dpy) ) { return VA_STATUS_ERROR_INVALID_DISPLAY; }

void va_TracePutSurface (
    VADisplay dpy,
    VASurfaceID surface,
    void *draw, /* the target Drawable */
    short srcx,
    short srcy,
    unsigned short srcw,
    unsigned short srch,
    short destx,
    short desty,
    unsigned short destw,
    unsigned short desth,
    VARectangle *cliprects, /* client supplied clip list */
    unsigned int number_cliprects, /* number of clip rects in the clip list */
    unsigned int flags /* de-interlacing flags */
);


VAStatus vaPutSurface (
    VADisplay dpy,
    VASurfaceID surface,
    Drawable draw, /* X Drawable */
    short srcx,
    short srcy,
    unsigned short srcw,
    unsigned short srch,
    short destx,
    short desty,
    unsigned short destw,
    unsigned short desth,
    VARectangle *cliprects, /* client supplied clip list */
    unsigned int number_cliprects, /* number of clip rects in the clip list */
    unsigned int flags /* de-interlacing flags */
)
{
  VADriverContextP ctx;

  if (fool_postp)
      return VA_STATUS_SUCCESS;

  CHECK_DISPLAY(dpy);
  ctx = CTX(dpy);
  
  VA_TRACE_FUNC(va_TracePutSurface, dpy, surface, (void *)draw, srcx, srcy, srcw, srch,
                destx, desty, destw, desth,
           cliprects, number_cliprects, flags );
  
  return ctx->vtable->vaPutSurface( ctx, surface, (void *)draw, srcx, srcy, srcw, srch,
                                   destx, desty, destw, desth,
                                   cliprects, number_cliprects, flags );
}
